
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

// ----------------------------------------------------------------------------

static std::string fileName = "";

// ----------------------------------------------------------------------------

void trim(std::string& str) {

	// find trailing whitespaces
	std::string::size_type pos = str.find_last_not_of(" \t\n\r");
	if (pos != std::string::npos) {
		// remove trailing whitespaces
		str.erase(pos+1);
		// find and remove leading whitespaces
		pos = str.find_first_not_of(" \t\n\r");
		if (pos != std::string::npos) {
			str.erase(0,pos);
		}
	} else {
		// only whitespaces -> clear string
		str.erase(str.begin(), str.end());
	}
}

// ----------------------------------------------------------------------------
void writeStreamToCout(std::istream* stream) {

	std::string str = "";

	while ((*stream).eof() == false) {
		getline((*stream),str);
		std::cout << str << std::endl;
	}
}

// ----------------------------------------------------------------------------
std::istream* DisableCodeWithCommentsFilter(std::istream* stream) {

    // * puts "//" in front of every code line

	std::string result = "";

	std::string str = "";

	while((*stream).eof() == false) {

		getline((*stream),str);
		result += "// "+str+"\n";
    }

	// create and return string stream
	delete stream;
	std::istringstream* resultStringStream = new std::istringstream(result);
	return resultStringStream;
}

// ----------------------------------------------------------------------------
std::istream* CFileFilter(std::istream* stream) {

    // * deals with forward declarations

	std::string result = "";

	std::string str = "";
	std::string strTrimmed = "";
	std::string currentBlock = "";
	std::string::size_type pos = std::string::npos;

	bool commentFlag = false;
	bool functionBody = false;
	bool define = false;
	bool newBlock = false;
	int curlyBrackets = 0;
	int definesCounter = 0;

	while((*stream).eof() == false) {

		getline((*stream),str);

        strTrimmed = str;
        trim(strTrimmed);

		// begin of comment block
		pos = str.find("/*++",0);
		if (pos != std::string::npos) {
			commentFlag = true;	// open comment
		}

		// end of comment block
		if (commentFlag == true) {
			pos = str.find("--*/",0);
			if (pos != std::string::npos) {
				commentFlag = false; // close comment
			}
		}

        // find function body
        if (commentFlag == false) {

            newBlock = (curlyBrackets == 0);

            if (functionBody == true) {

				if (str.find("#if") != std::string::npos) {
					definesCounter++;
				}

                if (str.find("#else") != std::string::npos) {
                    define = true;
                } else {
                    if (str.find("#endif") != std::string::npos) {
						definesCounter--;
						define = (definesCounter == 0);
                    }
                }
            }

            if (define == false) {
                // count opening {
                pos = str.find("{",0);
                while (pos != std::string::npos) {
                    curlyBrackets++;
                    pos = str.find("{",pos+1);
                }

                // count closing }
                pos = str.find("}",0);
                while (pos != std::string::npos) {
                    curlyBrackets--;
                    pos = str.find("}",pos+1);
                }

                // function body?
                functionBody = (curlyBrackets != 0);
            }

            newBlock = newBlock && functionBody;
        }

        if (((commentFlag == false) && (functionBody == false) && (strTrimmed == "")) || (newBlock == true)) {

            bool forwardDeclaration = false;

            pos = currentBlock.find("(",0);
            if (pos != std::string::npos) {
                if (currentBlock.find("(",pos+1) == std::string::npos) {
                    pos = currentBlock.find(");",0);
                    if (pos != std::string::npos) {
                        if (currentBlock.find(");",pos+1) == std::string::npos) {
                            forwardDeclaration = true;
                        }
                    }
                }
            }

            if (forwardDeclaration == false) {
                result += currentBlock;
            }

            currentBlock = "";
            newBlock = false;
        }

        if ((commentFlag == true) || (functionBody == true)) {
            result += str+"\n";
        } else {
            currentBlock += str+"\n";
        }
    }

	// create and return string stream
	delete stream;
	std::istringstream* resultStringStream = new std::istringstream(result);
	return resultStringStream;
}

// ----------------------------------------------------------------------------
std::istream* ASMFileFilter(std::istream* stream) {

	// * converts ASM-comments to Doxygen-comments
	// * extracts C function header for asm functions
	// * comments out all asm code lines

	std::string result = "";

	std::string str = "";
	std::string str_commentbuffer = "";
	std::string::size_type pos = std::string::npos;

	unsigned int routineState = 0;

	bool commentFlag = false;
	bool routineFlag = false;
	bool createDummyHeader = true;

	while((*stream).eof() == false) {

		getline((*stream),str);

		// begin of comment block
		pos = str.find(";++",0);
		if (pos != std::string::npos) {
			commentFlag = true;		// open comment
			str = "/*++";			// replace asm comment

			result += str+"\n";

			continue;
		}

		// end of comment block
		pos = str.find(";--",0);
		if (pos != std::string::npos) {

			str = "--*/";

			// found routine?
			if (routineFlag == false) {
				// write comment content
				result += str_commentbuffer;

				createDummyHeader = true;

			} else {
				// append routine header
				str += "\n"+str_commentbuffer;
			}

			commentFlag = false; // close comment
			routineFlag = false;
			str_commentbuffer = "";
		}

		if (commentFlag == true) {

			// remove leading ';'
			pos = str.find(";",0);
			if (pos != std::string::npos) {
				str.replace(pos,1,"");
			}

			// search for routines
			pos = str.find("Routine Description");
			if (pos != std::string::npos) {

				routineFlag = true;

				std::string dummy = str_commentbuffer;
				trim(dummy);

				routineState = 1;

				if (dummy != "") {
					createDummyHeader = false;
					// remove trailing ";"
					pos = str_commentbuffer.find(";");
					if (pos != std::string::npos) {
						str_commentbuffer.replace(pos,1,"");
					}
                    // HACK - i386/ctxswap.asm
                    if (str_commentbuffer.find("/*++") != std::string::npos) {
                        // remove!
                        str_commentbuffer.erase(str_commentbuffer.find(")")+1);
                    }
                    // ----
				}

			}

			// append to buffer
			if (routineFlag == false) {
				str_commentbuffer += str+"\n";
			}

		} else {

			if (str.find("--*/") == std::string::npos) {
				str = "// "+str;
			}

			if (routineState == 0) {

				pos = str.find("cPublicProc");
				if (pos != std::string::npos) {

					if (createDummyHeader == true) {
						std::string routineName = str.substr(pos+12,str.find(",")-pos-12);
						result = result + "void "+routineName+"( /*???*/ )\n";

					} else createDummyHeader = true;

					str = "{";
					routineState = 2;
				}

				pos = str.find("NESTED_ENTRY");
				if (pos != std::string::npos) {

					if (createDummyHeader == true) {
						std::string routineName = str.substr(pos+13,str.find(",")-pos-13);
						result = result + "void "+routineName+"( /*???*/ )\n";

					} else createDummyHeader = true;

					str = "{";
					routineState = 2;
				}
			}

			if (routineState == 1) {

				// --- HACK for i386/ctxswap.asm --------------------
				if ((fileName.find("ctxswap.asm") != std::string::npos) &&
                    (fileName.find("i386") != std::string::npos)) {

					pos = str.find("SwapContext     proc");
					if (pos != std::string::npos) {
						result = result+"void SwapContext( /*???*/ )\n";
						createDummyHeader = true;
						str = "{";
						routineState = 2;
					}
				}
				// --- HACK -----------------------------------------

				pos = str.find("cPublicProc");
				if (pos != std::string::npos) {
					if (createDummyHeader == true) {
						std::string routineName = str.substr(pos+12,str.find(",")-pos-12);
						result = result + "void "+routineName+"( /*???*/ )\n";

					} else createDummyHeader = true;

					str = "{";
					routineState = 2;
				}

				pos = str.find("NESTED_ENTRY");
				if (pos != std::string::npos) {

					if (createDummyHeader == true) {
						std::string routineName = str.substr(pos+13,str.find(",")-pos-13);
						result = result + "void "+routineName+"( /*???*/ )\n";

					} else createDummyHeader = true;

					str = "{";
					routineState = 2;
				}

				pos = str.find("cPublicFastCall");
				if (pos != std::string::npos) {

					str = "{";
					routineState = 2;
				}
			}

			if (routineState == 2) {

				// --- HACK for i386/ctxswap.asm --------------------
				if ((fileName.find("ctxswap.asm") != std::string::npos) &&
                    (fileName.find("i386") != std::string::npos)) {

                    pos = str.find("SwapContext     endp");
					if (pos != std::string::npos) {
						str = "}";
						routineState = 0;
					}
				}
				// --- HACK -----------------------------------------

				pos = str.find("stdENDP");
				if (pos != std::string::npos) {
					str = "}";
					routineState = 0;
				}

				pos = str.find("NESTED_END");
				if (pos != std::string::npos) {
					str = "}";
					routineState = 0;
				}

				pos = str.find("fstENDP");
				if (pos != std::string::npos) {
					str = "}";
					routineState = 0;
				}

				createDummyHeader = true;
			}
		}

		if ((commentFlag == false) || (routineFlag == true)) {
			result += str+"\n";
		}
	}

	// create and return string stream
	delete stream;
	std::istringstream* resultStringStream = new std::istringstream(result);
	return resultStringStream;
}

// ----------------------------------------------------------------------------
std::istream* DefaultFileFilter(std::istream* stream) {

	// * converts WRK-comments to Doxygen-comments
	// * converts WRK-parameter/returnvalue descriptions to Doxygen-form
	// * finds and marks file/module description
	// * inserts link to "License.txt" file

	std::string result = "";

	std::string str = "";
	std::string str_trimmed = "";
	std::string::size_type pos = std::string::npos;

	unsigned int replaceCount = 0;

	unsigned int state = 0;

	bool startFlag = true;
	bool commentFlag = false;

	while((*stream).eof() == false) {

		getline((*stream),str);
		str_trimmed = str;
		trim(str_trimmed);

		replaceCount = 4;
		pos = str.find("/*++",0);
		// HACK
		if (pos == std::string::npos) {
			replaceCount = 5;
			pos = str.find("/**++");
		}
		// HACK

		if (pos != std::string::npos) {

			commentFlag = true; // open comment

			//str.replace(pos,replaceCount,"/*!");
			str = "/*!";

			if (startFlag == true) {
				// first comment block -> file header
				str += "\n\\file";
			}
		}

		replaceCount = 4;
		pos = str.find("--*/",0);
		if (pos == std::string::npos) {
			replaceCount = 3;
			pos = str.find("-*/");
		}

		if (pos != std::string::npos) {

			startFlag = false;

			commentFlag = false; // close comment
			state = 0;

			str.replace(pos,replaceCount,"*/");
		}

		if (commentFlag == true) {

			bool foundSpecialSection = false;

			pos = str.find(":");
			if (pos != std::string::npos) {

				if (str.find("Routine Description:") != std::string::npos) {
					state = 1;
					str = "";
					foundSpecialSection = true;
				} else
					if ((str.find("Arguments:") != std::string::npos) ||
						(str.find("Parameters:") != std::string::npos)) {
						state = 2;
						str = "";
						foundSpecialSection = true;
					} else 
						if ((str.find("Return Value:") != std::string::npos) ||
							(str.find("Return value:") != std::string::npos) ||
							(str.find("Returned Value:") != std::string::npos)) {
							state = 3;
							str = "";
							foundSpecialSection = true;
						} else {
							if ((str.find("Abstract:") != std::string::npos) ||
								(str.find("Module Name:") != std::string::npos) ||
								(str.find("Environment:") != std::string::npos) ||
								(str.find("Revision History:") != std::string::npos)) {

                                // HACK - replace ":" with HTML definition
                                pos = str_trimmed.find(":");
                                while (pos != std::string::npos) {
                                    str_trimmed.replace(pos,1,"\\htmlonly&#58;\\endhtmlonly");
                                    pos = str_trimmed.find(":");
                                }
                                // HACK

								str = "<em>"+str_trimmed+"</em>";
								foundSpecialSection = true;

							    state = 0;
							}
						}

			}

			if (foundSpecialSection == false) {

				pos = str.find("License.txt");
				if (pos != std::string::npos) {
					str.replace(pos,11,"<a href=\"../../../license.txt\">License.txt</a>");
				}

				if (str_trimmed != "") {
					switch ( state ) {
						case 1: { 
									break;
								}
						case 2: {
									if (str_trimmed.find("None") == std::string::npos) {
											if (str.find("-") != std::string::npos) {
												str = "\\param "+str;
											}
									} else {
										str = "";
									}
									break;
								}
						case 3: { 
									str = "\\return "+str;
									break;
								}
					}
				}

                // HACK - replace ":" with HTML definition
                pos = str.find(":");
                while (pos != std::string::npos) {
                    str.replace(pos,1,"\\htmlonly&#58;\\endhtmlonly");
                    pos = str.find(":");
                }
                // HACK
			}
		}

		result += str+"\n";
	}

	// create and return string stream
	delete stream;
	std::istringstream* resultStringStream = new std::istringstream(result);
	return resultStringStream;
}

// ----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	// get current file name
	fileName = argv[1];

	// open file stream
	std::ifstream* inputFile = new std::ifstream(fileName.c_str());

	// #### HACK ####
	if ((fileName.substr(fileName.length()-8,8) == "trap.asm") &&
		(fileName.find("i386") != std::string::npos)) {

		writeStreamToCout(DisableCodeWithCommentsFilter(inputFile));
	}
	// #### ####

	// check file type
	if (fileName.substr(fileName.length()-4,4) == ".asm") {

		// ASM filter
		writeStreamToCout(DefaultFileFilter(ASMFileFilter(inputFile)));
//		writeStreamToCout(ASMFileFilter(inputFile));

	} else

    if (fileName.substr(fileName.length()-2,2) == ".c") {

        // C filter
        writeStreamToCout(DefaultFileFilter(CFileFilter(inputFile)));
//        writeStreamToCout(CFileFilter(inputFile));

    } else

	{
		// default filter
		writeStreamToCout(DefaultFileFilter(inputFile));
	}

	return 0;
}
